@@ -37,7 +37,7 @@ def cast_int(x):
     return int(x)
 
 
-construct_bytes = bytes
+make_bytes = bytes
 
 dtype_from_string = {
     "float32": torch.float32,
