@@ -14,13 +14,13 @@ use itertools::Itertools;
 use crate::{all_imports::RInts, pyo3_prelude::*, sv, util::BitMask128};
 
 use super::{
-    computational_node::EinsumArgs, deep_map_unwrap, evaluate_fn, prelude::*, Add, Concat, Einsum,
-    GeneralFunction, HashBytes, Index, Rearrange, ScalarConstant, Scatter, Symbol,
+    computational_node::EinsumArgs, deep_map_unwrap, evaluate_fn, prelude::*,
+    py_circuit_items::PY_CIRCUIT_ITEMS, Add, Concat, Einsum, GeneralFunction, HashBytes, Index,
+    Rearrange, ScalarConstant, Scatter, Symbol,
 };
 use crate::{
     filter_by_variant,
     opt_einsum::optimize_einsum_spec_cached,
-    py_types::PY_CIRC_COMPILER_UTIL,
     rearrange_spec::{shape_to_op_shape, RearrangeSpec},
     tensor_util::{
         compose, Shape, Slice, TensorAxisIndex, TensorIndex, TensorInvariantError, USlice,
@@ -1261,7 +1261,8 @@ pub fn push_down_index(node: &Index) -> Option<CircuitRc> {
         Circuit::Rearrange(inner) => {
             // doing this in python instead of porting to rust bc it has a lot of torch ops so likely wouldn't be faster in rust
             let python_result: Option<(TensorIndex, RearrangeSpec)> = Python::with_gil(|py| {
-                PY_CIRC_COMPILER_UTIL
+                PY_CIRCUIT_ITEMS
+                    .circ_compiler_util
                     .getattr(py, "index_before_rearrange_rust")
                     .unwrap()
                     .call(
