@@ -1,8 +1,20 @@
 #![allow(clippy::borrow_deref_ref)]
 use crate::hashmaps::{AHashSet as HashSet, FxHashMap as HashMap};
-use crate::py_types::PY_MAKE_BYTES;
-use crate::{pycall, pyo3_prelude::*};
+use crate::pyo3_prelude::*;
+use crate::{
+    cached_lambda,
+    py_types::{make_bytes_py, PyShape, Tensor},
+    rearrange_spec::{OpSize, RearrangeSpec},
+    tensor_util::{Shape, TensorAxisIndex, TensorIndex, TorchDeviceDtype, TorchDeviceDtypeOp},
+    util::AsOp,
+};
+pub use computational_node::{
+    flat_concat, Add, Concat, Einsum, EinsumAxes, Index, Rearrange, Scatter,
+};
+pub use constant::{ArrayConstant, ScalarConstant, Symbol};
+pub use generalfunction::{GeneralFunction, GeneralFunctionSpec, PyGFSpecShapeGetter};
 use num_bigint::BigUint;
+use py_circuit_items::circuit_rust_to_py;
 use pyo3::{exceptions, pyclass::CompareOp};
 use std::hash::Hash;
 use std::{
@@ -10,6 +22,7 @@ use std::{
     ops::{Deref, DerefMut},
     rc::Rc,
 };
+use uuid::uuid;
 
 use macro_rules_attribute::apply;
 use std::fmt::Debug;
@@ -33,26 +46,12 @@ pub mod generalfunction;
 pub mod named_axes;
 pub mod nrc;
 pub mod print;
+pub mod py_circuit_items;
 mod repr;
 pub mod scatter_rewrite;
 pub mod scheduled_execution;
 pub mod scheduling_z3;
 
-use crate::{
-    cached_lambda,
-    py_types::{PyShape, Tensor, CIRCUIT_RUST_TO_PY},
-    rearrange_spec::{OpSize, RearrangeSpec},
-    tensor_util::{Shape, TensorAxisIndex, TensorIndex, TorchDeviceDtype, TorchDeviceDtypeOp},
-    util::AsOp,
-};
-pub use computational_node::{
-    flat_concat, Add, Concat, Einsum, EinsumAxes, Index, Rearrange, Scatter,
-};
-pub use generalfunction::{GeneralFunction, GeneralFunctionSpec, PyGFSpecShapeGetter};
-
-pub use constant::{ArrayConstant, ScalarConstant, Symbol};
-use uuid::uuid;
-
 mod circuit_node_private {
     pub trait CircuitNodePrivate {
         fn info_mut(&mut self) -> &mut super::CachedCircuitInfo;
@@ -705,8 +704,8 @@ pub type HashBytes = [u8; 32];
 #[derive(Clone, Copy, Hash, PartialEq, Eq)]
 pub struct PyHashBytes(pub [u8; 32]);
 impl IntoPy<PyObject> for PyHashBytes {
-    fn into_py(self, _py: Python<'_>) -> PyObject {
-        pycall!(PY_MAKE_BYTES, (self.0,))
+    fn into_py(self, py: Python<'_>) -> PyObject {
+        make_bytes_py(py, &self.0).unwrap()
     }
 }
 
@@ -1202,10 +1201,6 @@ fn python_apply_fn_to_sub<N: CircuitNode>(
     })
 }
 
-pub fn circuit_rust_to_py(circ: CircuitRc) -> PyObject {
-    Python::with_gil(|py| CIRCUIT_RUST_TO_PY.call(py, (circ,), None).unwrap())
-}
-
 #[pymethods]
 impl PyCircuitBase {
     #[getter]
