@@ -1,5 +1,5 @@
+use crate::hashmaps::{AHashSet as HashSet, FxHashMap as HashMap};
 use std::cmp::max;
-use std::collections::{HashMap, HashSet};
 
 use z3;
 
