@@ -1,5 +1,6 @@
-use crate::pyo3_prelude::*;
-use lazy_static::lazy_static;
+use crate::{all_imports::TorchDeviceDtype, pyo3_prelude::*};
+use macro_rules_attribute::apply;
+use pyo3::types::IntoPyDict;
 use pyo3::{
     exceptions,
     types::{PyBytes, PyTuple},
@@ -8,104 +9,196 @@ use uuid::Uuid;
 
 use crate::{
     circuit::{EinsumAxes, HashBytes},
+    lazy::GILLazyPy,
     tensor_util::Shape,
 };
 
-lazy_static! {
-    static ref PYTHON_UTILS: Py<PyModule> = Python::with_gil(|py| {
-        PyModule::from_code(
-            py,
-            include_str!(concat!(env!("CARGO_MANIFEST_DIR"), "/rust_circuit_type_utils.py")),
-            "rust_circuit_type_utils.py",
-            "rust_circuit_type_utils",
-        )
-        .unwrap() // TODO: fail more gracefully?
-        .into()
-    });
-
-    pub static ref PY_TORCH: PyObject =
-        Python::with_gil(|py| PYTHON_UTILS.getattr(py, "torch").unwrap());
-    static ref GET_TENSOR_SHAPE: PyObject =
-        Python::with_gil(|py| PYTHON_UTILS.getattr(py, "get_tensor_shape").unwrap());
-    static ref GET_UUID_BYTES: PyObject =
-        Python::with_gil(|py| PYTHON_UTILS.getattr(py, "get_uuid_bytes").unwrap());
-    static ref CONSTRUCT_UUID_BYTES: PyObject =
-        Python::with_gil(|py| PYTHON_UTILS.getattr(py, "construct_uuid_bytes").unwrap());
-    static ref PY_OPERATOR: Py<PyModule> =
-        Python::with_gil(|py| PyModule::import(py, "operator").unwrap().into());
-    pub static ref PY_ID: PyObject =
-        Python::with_gil(|py| PYTHON_UTILS.getattr(py, "get_id").unwrap());
-    pub static ref PY_CAST_INT: PyObject =
-        Python::with_gil(|py| PYTHON_UTILS.getattr(py, "cast_int").unwrap());
-    pub static ref PY_SCALAR_TO_TENSOR: PyObject =
-        Python::with_gil(|py| PYTHON_UTILS.getattr(py, "scalar_to_tensor").unwrap());
-    pub static ref CAST_TENSOR: PyObject =
-        Python::with_gil(|py| PYTHON_UTILS.getattr(py, "cast_tensor").unwrap());
-    pub static ref UN_FLAT_CONCAT: PyObject =
-        Python::with_gil(|py| PYTHON_UTILS.getattr(py, "un_flat_concat").unwrap());
-    pub static ref TENSOR_SCALE: PyObject =
-        Python::with_gil(|py| PYTHON_UTILS.getattr(py, "tensor_scale").unwrap());
-    pub static ref GENERALFUNCTIONS: std::collections::HashMap<String,PyObject> =
-        Python::with_gil(|py| PYTHON_UTILS.getattr(py, "generalfunctions").unwrap().extract(py).unwrap());
-    pub static ref PY_NONE: PyObject=
-        Python::with_gil(|py| PYTHON_UTILS.getattr(py, "none").unwrap().extract(py).unwrap());
-    pub static ref PY_EINSUM: PyObject=
-        Python::with_gil(|py| PYTHON_UTILS.getattr(py, "einsum").unwrap().extract(py).unwrap());
-    pub static ref PY_MAKE_DIAGONAL: PyObject=
-        Python::with_gil(|py| PYTHON_UTILS.getattr(py, "make_diagonal").unwrap().extract(py).unwrap());
-    pub static ref PY_MAKE_BYTES: PyObject=
-        Python::with_gil(|py| PYTHON_UTILS.getattr(py, "construct_bytes").unwrap().extract(py).unwrap());
+pub struct PyUtils {
+    pub torch: PyObject,
+    get_tensor_shape: PyObject,
+    get_uuid_bytes: PyObject,
+    construct_uuid_bytes: PyObject,
+    id: PyObject,
+    pub cast_int: PyObject,
+    scalar_to_tensor: PyObject,
+    pub cast_tensor: PyObject,
+    un_flat_concat: PyObject,
+    tensor_scale: PyObject,
+    pub generalfunctions: std::collections::HashMap<String, PyObject>,
+    pub none: PyObject,
+    einsum: PyObject,
+    make_diagonal: PyObject,
+    make_bytes: PyObject,
 }
 
-lazy_static! {
-    static ref PYCIRCUIT_MODULE: Py<PyModule> = Python::with_gil(|py| {
-        PyModule::from_code(
-            py,
-            "from interp.circuit import computational_node, constant; import interp.circuit.interop_rust.interop_rust as interop; import einops; import interp.circuit.circuit_compiler.util as circ_compiler_util",
-            "py_circuit_interop.py",
-            "py_circuit_interop",
-        )
-        .unwrap() // TODO: fail more gracefully?
-        .into()
-    });
+/// misc python utilities
+pub static PY_UTILS: GILLazyPy<PyUtils> = GILLazyPy::new_py(|py| {
+    let utils = PyModule::from_code(
+        py,
+        include_str!(concat!(
+            env!("CARGO_MANIFEST_DIR"),
+            "/rust_circuit_type_utils.py"
+        )),
+        "rust_circuit_type_utils.py",
+        "rust_circuit_type_utils",
+    )
+    .unwrap();
 
-    pub static ref PY_COMPUTATIONAL_NODE: PyObject =
-        Python::with_gil(|py| PYCIRCUIT_MODULE.getattr(py, "computational_node").unwrap());
-    pub static ref PY_CONSTANT: PyObject =
-        Python::with_gil(|py| PYCIRCUIT_MODULE.getattr(py, "constant").unwrap());
-    pub static ref PY_CIRCUIT_INTEROP: PyObject =
-        Python::with_gil(|py| PYCIRCUIT_MODULE.getattr(py, "interop").unwrap());
-    pub static ref CIRCUIT_RUST_TO_PY: PyObject =
-        Python::with_gil(|py| PY_CIRCUIT_INTEROP.getattr(py, "rust_to_py").unwrap());
-    pub static ref PY_EINOPS: PyObject =
-        Python::with_gil(|py| PYCIRCUIT_MODULE.getattr(py, "einops").unwrap());
-    pub static ref PY_EINOPS_REPEAT: PyObject =
-        Python::with_gil(|py| PY_EINOPS.getattr(py, "repeat").unwrap());
-    pub static ref PY_CIRC_COMPILER_UTIL: PyObject =
-        Python::with_gil(|py| PYCIRCUIT_MODULE.getattr(py, "circ_compiler_util").unwrap());
+    let get = |s: &str| utils.getattr(s).unwrap().into();
 
-    static ref PY_TENSOR_HASH_MODULE: Py<PyModule> = Python::with_gil(|py| {
-        PyModule::from_code(
-            py,
-            include_str!(concat!(env!("CARGO_MANIFEST_DIR"), "/tensor_hash.py")),
-            "tensor_hash.py",
-            "tensor_hash",
-        )
-        .unwrap() // TODO: fail more gracefully?
-        .into()
-    });
+    PyUtils {
+        torch: get("torch"),
+        get_tensor_shape: get("get_tensor_shape"),
+        get_uuid_bytes: get("get_uuid_bytes"),
+        construct_uuid_bytes: get("construct_uuid_bytes"),
+        id: get("get_id"),
+        cast_int: get("cast_int"),
+        scalar_to_tensor: get("scalar_to_tensor"),
+        cast_tensor: get("cast_tensor"),
+        un_flat_concat: get("un_flat_concat"),
+        tensor_scale: get("tensor_scale"),
+        generalfunctions: utils
+            .getattr("generalfunctions")
+            .unwrap()
+            .extract()
+            .unwrap(),
+        none: get("none"),
+        einsum: get("einsum"),
+        make_diagonal: get("make_diagonal"),
+        make_bytes: get("make_bytes"),
+    }
+});
 
-
-    pub static ref HASH_TENSOR: PyObject =
-        Python::with_gil(|py| PY_TENSOR_HASH_MODULE.getattr(py, "hash_tensor").unwrap());
+pub fn py_address(x: &PyObject) -> usize {
+    Python::with_gil(|py| {
+        PY_UTILS
+            .id
+            .call(py, (x,), None)
+            .unwrap()
+            .extract(py)
+            .unwrap()
+    })
 }
 
+#[macro_export]
+macro_rules! make_py_func {
+    {
+        #[py_ident($py:ident)]
+        $( #[$m:meta] )*
+        $vi:vis fn $name:ident($($arg_name:ident : $arg_ty:ty),* $(,)?) -> $ret_ty:ty
+        { $($tt:tt)* }
+    } => {
+        paste::paste!{
+            $(#[$m])*
+            $vi fn [<$name _py>]<'py>($py : Python<'py>, $($arg_name : $arg_ty,)*) -> $ret_ty {
+                $($tt)*
+            }
+            $(#[$m])* // TODO: maybe shouldn't apply to both??
+            $vi fn $name($($arg_name : $arg_ty,)*) -> $ret_ty {
+                Python::with_gil(|py| [<$name _py>](py, $($arg_name,)*))
+            }
+        }
+
+    };
+}
+
+#[apply(make_py_func)]
+#[py_ident(py)]
+pub fn scalar_to_tensor(v: f64, shape: Shape, device_dtype: TorchDeviceDtype) -> PyResult<Tensor> {
+    PY_UTILS
+        .scalar_to_tensor
+        .call(py, (v, shape, device_dtype), None)?
+        .extract::<Tensor>(py)
+}
+
+#[apply(make_py_func)]
+#[py_ident(py)]
+pub fn einsum(items: Vec<(Tensor, EinsumAxes)>, out_axes: EinsumAxes) -> PyResult<Tensor> {
+    PY_UTILS
+        .einsum
+        .call(py, (items, out_axes), None)?
+        .extract(py)
+}
+
+#[apply(make_py_func)]
+#[py_ident(py)]
+pub fn make_diagonal(
+    non_diag: &Tensor,
+    out_axes_deduped: EinsumAxes,
+    out_axes: EinsumAxes,
+) -> PyResult<Tensor> {
+    PY_UTILS
+        .make_diagonal
+        .call(py, (non_diag, out_axes_deduped, out_axes), None)?
+        .extract(py)
+}
+
+#[apply(make_py_func)]
+#[py_ident(py)]
+pub fn einops_repeat(
+    tensor: &Tensor,
+    op: String,
+    sizes: impl IntoIterator<Item = (String, u64)>,
+) -> PyResult<Tensor> {
+    PY_EINOPS
+        .repeat
+        .call(py, (tensor, op), Some(sizes.into_py_dict(py)))?
+        .extract(py)
+}
+
+#[apply(make_py_func)]
+#[py_ident(py)]
+pub fn make_bytes(bytes: &[u8]) -> PyResult<PyObject> {
+    PY_UTILS.make_bytes.call(py, (bytes,), None)
+}
+
+#[apply(make_py_func)]
+#[py_ident(py)]
+pub fn un_flat_concat(tensor: &Tensor, split_shapes: Vec<Shape>) -> PyResult<Vec<Tensor>> {
+    PY_UTILS
+        .un_flat_concat
+        .call(py, (tensor, split_shapes), None)?
+        .extract(py)
+}
+
+#[apply(make_py_func)]
+#[py_ident(py)]
+pub fn tensor_scale(tensor: &Tensor) -> PyResult<f64> {
+    PY_UTILS.tensor_scale.call(py, (tensor,), None)?.extract(py)
+}
+
+pub struct PyEinops {
+    pub einops: Py<PyModule>,
+    pub repeat: PyObject,
+}
+
+pub static PY_EINOPS: GILLazyPy<PyEinops> = GILLazyPy::new_py(|py| {
+    let einops = PyModule::import(py, "einops").unwrap();
+    let get = |s: &str| einops.getattr(s).unwrap().into();
+
+    PyEinops {
+        einops: einops.into(),
+        repeat: get("repeat"),
+    }
+});
+
+pub static HASH_TENSOR: GILLazyPy<PyObject> = GILLazyPy::new_py(|py| {
+    let module = PyModule::from_code(
+        py,
+        include_str!(concat!(env!("CARGO_MANIFEST_DIR"), "/tensor_hash.py")),
+        "tensor_hash.py",
+        "tensor_hash",
+    )
+    .unwrap();
+    module.getattr("hash_tensor").unwrap().into()
+});
+
 #[macro_export]
 macro_rules! pycall {
-    ($f:ident,$args:expr) => {
+    ($f:expr,$args:expr) => {
         Python::with_gil(|py| $f.call(py, $args, None).unwrap().extract(py).unwrap())
     };
-    ($f:ident,$args:expr,$err_type:ty) => {
+    ($f:expr,$args:expr,$err_type:ty) => {
         Python::with_gil(|py| {
             $f.call(py, $args, None)
                 .map_err(|z| <$err_type>::from(z))?
@@ -118,13 +211,19 @@ macro_rules! pycall {
 macro_rules! generate_extra_py_ops {
     [$($op:ident),*] => {
         paste::paste! {
-            lazy_static! {
-                $(
-                    static ref [<PY_ $op:upper>]: PyObject =
-                        Python::with_gil(|py| PY_OPERATOR.getattr(py, stringify!($op)).unwrap());
-                )*
+            struct PyOperators {
+                $($op: PyObject,)*
             }
 
+            static PY_OPERATORS: GILLazyPy<PyOperators> = GILLazyPy::new_py(|py| {
+                let operator = PyModule::import(py, "operator").unwrap();
+
+                PyOperators {
+                    $( $op : operator.getattr(stringify!($op)).unwrap().into(),)*
+                }
+            });
+
+
             /// Trait for python operator methods when they return the same type.
             /// Used for tensors.
             ///
@@ -166,7 +265,7 @@ macro_rules! generate_extra_py_ops {
             {
                 $(
                     fn [<py_ $op>]<'a>(self, py: Python<'a>, x: impl IntoPy<PyObject>) -> PyResult<Self> {
-                        [<PY_ $op:upper>].call1(py, (self, x))?.extract(py)
+                        PY_OPERATORS.$op.call1(py, (self, x))?.extract(py)
                     }
 
                     // not sure if this method should exist
@@ -187,7 +286,7 @@ pub struct PyUuid(pub Uuid);
 impl<'source> FromPyObject<'source> for PyUuid {
     fn extract(uuid_obj: &'source PyAny) -> PyResult<Self> {
         let uuid_bytes: Vec<u8> =
-            Python::with_gil(|py| GET_UUID_BYTES.call1(py, (uuid_obj,))?.extract(py))?;
+            Python::with_gil(|py| PY_UTILS.get_uuid_bytes.call1(py, (uuid_obj,))?.extract(py))?;
 
         let num_bytes = uuid_bytes.len();
         let bytes_arr: [u8; 16] = uuid_bytes.try_into().map_err(|_| {
@@ -203,7 +302,8 @@ impl<'source> FromPyObject<'source> for PyUuid {
 
 impl IntoPy<PyObject> for PyUuid {
     fn into_py(self, py: Python<'_>) -> PyObject {
-        CONSTRUCT_UUID_BYTES
+        PY_UTILS
+            .construct_uuid_bytes
             .call1(py, (PyBytes::new(py, &self.0.into_bytes()),))
             .unwrap()
     }
@@ -218,7 +318,8 @@ pub struct Tensor {
 
 impl<'source> FromPyObject<'source> for Tensor {
     fn extract(tensor: &'source PyAny) -> PyResult<Self> {
-        let shape = Python::with_gil(|py| GET_TENSOR_SHAPE.call1(py, (tensor,))?.extract(py))?;
+        let shape =
+            Python::with_gil(|py| PY_UTILS.get_tensor_shape.call1(py, (tensor,))?.extract(py))?;
 
         Ok(Self {
             tensor: tensor.into(),
@@ -228,6 +329,11 @@ impl<'source> FromPyObject<'source> for Tensor {
     }
 }
 
+impl IntoPy<PyObject> for &Tensor {
+    fn into_py(self, _py: Python<'_>) -> PyObject {
+        self.tensor.clone()
+    }
+}
 impl IntoPy<PyObject> for Tensor {
     fn into_py(self, _py: Python<'_>) -> PyObject {
         self.tensor
@@ -293,7 +399,3 @@ impl IntoPy<PyObject> for PyEinsumAxes {
         PyTuple::new(py, self.0).into_py(py)
     }
 }
-
-pub fn py_address(x: &PyObject) -> usize {
-    Python::with_gil(|py| PY_ID.call(py, (x,), None).unwrap().extract(py).unwrap())
-}
