@@ -191,7 +191,7 @@ impl USlice {
 
     pub fn shrink_base(&self, new_base: &USlice) -> USlice {
         assert!(new_base.stop >= self.stop);
-        assert!(self.start >= new_base.start && self.stop >= new_base.stop);
+        assert!(self.start >= new_base.start);
         USlice {
             start: self.start - new_base.start,
             stop: self.stop - new_base.start,
