@@ -1,5 +1,9 @@
-use crate::hashmaps::{AHashSet as HashSet, FxHashMap as HashMap};
-use lazy_static::lazy_static;
+use crate::{
+    hashmaps::{AHashSet as HashSet, FxHashMap as HashMap},
+    make_lazy,
+};
+use macro_rules_attribute::apply;
+use once_cell::sync::Lazy;
 
 use std::hash::Hash;
 
@@ -8,6 +12,8 @@ pub type EinInt = u8;
 /// letters used in einsum and rearrange strings
 /// i wanted this to be a constant, but that is annoying in rust bc can't do any computation to produce constants
 /// so instead it's cached function
+#[apply(make_lazy)]
+#[lazy_ty(Lazy)]
 pub fn alphabet() -> Vec<String> {
     let alphabet_upper: Vec<String> = (b'A'..=b'Z')
         .map(|x| std::str::from_utf8(&[x]).unwrap().to_owned())
@@ -21,14 +27,16 @@ pub fn alphabet() -> Vec<String> {
         .cloned()
         .collect()
 }
-lazy_static! {
-    pub static ref ALPHABET: Vec<String> = alphabet();
-    pub static ref ALPHABET_INV: HashMap<String, usize> = ALPHABET
+
+#[apply(make_lazy)]
+#[lazy_ty(Lazy)]
+pub fn alphabet_inv() -> HashMap<String, usize> {
+    ALPHABET
         .iter()
         .cloned()
         .enumerate()
         .map(|(a, b)| (b, a))
-        .collect();
+        .collect()
 }
 
 pub fn is_unique<T: Eq + Hash>(col: &[T]) -> bool {
